import os

from hyfi import HyFI

from ._version import __version__

# Read the package name from the current directory
__package_path__ = os.path.dirname(__file__)

# Initialize the global HyFI object
HyFI.initialize_global_hyfi(
    package_path=__package_path__,
    version=__version__,
    plugins=[],
)

# Initialize the logger
HyFI.setLogger()


def get_version() -> str:
    """Get the package version."""
    return __version__


__all__ = ["HyFI", "get_version"]



from typing import Any, List, Optional, Tuple

from pydantic import BaseModel, model_validator

from lexikanon import HyFI
from lexikanon.tokenizers.base import Tokenizer

logger = HyFI.getLogger(__name__)


class NLTKTagger(BaseModel):
    """
    lemmatize: false
    stem: true
    lemmatizer:
        _target_: nltk.stem.WordNetLemmatizer
    stemmer:
        _target_: nltk.stem.PorterStemmer
    """

    lemmatize: bool = False
    stem: bool = True
    lemmatizer: Optional[dict] = None
    stemmer: Optional[dict] = None
    verbose: bool = False

    _lemmatizer: Any = None
    _stemmer: Any = None

    @model_validator(mode="after")
    def validate_nltk(self) -> "NLTKTagger":
        import nltk as NLTK

        NLTK.download("punkt", quiet=True)
        NLTK.download("averaged_perceptron_tagger", quiet=True)
        NLTK.download("wordnet", quiet=True)
        NLTK.download("omw-1.4", quiet=True)

        if self.lemmatizer and HyFI.is_instantiatable(self.lemmatizer):
            logger.info("instantiating %s...", self.lemmatizer["_target_"])
            self._lemmatizer = HyFI.instantiate(self.lemmatizer)
        if self.stemmer and HyFI.is_instantiatable(self.stemmer):
            logger.info("instantiating %s...", self.stemmer["_target_"])
            self._stemmer = HyFI.instantiate(self.stemmer)
        self.lemmatize = self.lemmatize and self._lemmatizer is not None
        self.stem = self.stem and self._stemmer is not None

        return self

    def _parse(self, text: str) -> List[Tuple[str, str]]:
        import nltk

        tokens: List[tuple] = nltk.pos_tag(nltk.word_tokenize(text))
        return tokens

    def _lemmatize(self, token_pos: Tuple[str, str]) -> Tuple[str, str]:
        if self._lemmatizer is None:
            return token_pos
        return (
            self._lemmatizer.lemmatize(
                token_pos[0], self._get_wordnet_pos(token_pos[1])
            ),
            token_pos[1],
        )

    def _stem(self, token_pos: Tuple[str, str]) -> Tuple[str, str]:
        if self.stemmer is None:
            return token_pos
        return (self._stemmer.stem(token_pos[0]), token_pos[1])

    @staticmethod
    def _get_wordnet_pos(tag: str) -> str:
        from nltk.corpus import wordnet

        """Map POS tag to first character lemmatize() accepts"""
        tag = tag[0].upper()
        tag_dict = {
            "J": wordnet.ADJ,
            "N": wordnet.NOUN,
            "V": wordnet.VERB,
            "R": wordnet.ADV,
        }

        return tag_dict.get(tag, wordnet.NOUN)


class NLTKTokenizer(Tokenizer):
    tagger: NLTKTagger = NLTKTagger()

    def parse(self, text: str) -> List[Tuple[str, str]]:
        token_tuples = self.tagger._parse(text)
        tokens = []
        for token_tuple in token_tuples:
            if self.tagger.lemmatize:
                token_tuple = self.tagger._lemmatize(token_tuple)
            if self.tagger.stem:
                token_tuple = self.tagger._stem(token_tuple)
            tokens.append(self.to_token(token_tuple))
        return tokens



from typing import Callable, List, Optional, Union

from pydantic import BaseModel, model_validator

from lexikanon import HyFI

logger = HyFI.getLogger(__name__)


class Stopwords(BaseModel):
    """
    name: stopwords
    lowercase: true
    stopwords:
    stopwords_path:
    nltk_stopwords_lang:
    verbose: False
    """

    name: str = "stopwords"
    lowercase: bool = False
    stopwords: Optional[Union[List[str], Callable[[str], bool]]] = None
    stopwords_path: Optional[str] = None
    nltk_stopwords_lang: Optional[str] = None
    verbose: bool = False

    _stopwords_list: List[str] = []
    _stopwords_fn: Callable = lambda x: False

    @model_validator(mode="after")
    def validate_stopwords(self) -> "Stopwords":
        if self.stopwords_path:
            self._stopwords_list = HyFI.load_wordlist(
                self.stopwords_path, lowercase=self.lowercase, verbose=self.verbose
            )
            logger.info(
                "Loaded %d stopwords from %s",
                len(self._stopwords_list),
                self.stopwords_path,
            )
        else:
            self._stopwords_list = []

        if callable(self.stopwords):
            self._stopwords_fn = self.stopwords
            logger.info("Using custom stopwords function %s", self.stopwords)
        elif isinstance(self.stopwords, list):
            if self.lowercase:
                self.stopwords = [w.lower() for w in self.stopwords]
            logger.info("Loaded %d stopwords", len(self.stopwords))
            self._stopwords_list += self.stopwords

        if self.nltk_stopwords_lang:
            self._stopwords_list += self._load_nltk_stopwords(self.nltk_stopwords_lang)
        logger.info("Loaded %d stopwords", len(self._stopwords_list))
        return self

    def __call__(self, word: str) -> bool:
        """Calling a stopwords instance like a function just calls the is_stopword method."""
        return self.is_stopword(word)

    def is_stopword(self, word: str) -> bool:
        """
        :type word: str
        :returns: bool
        """
        _word = word.lower() if self.lowercase else word
        return self._stopwords_fn(_word) or _word in self._stopwords_list

    def _load_nltk_stopwords(self, language: str = "english") -> List[str]:
        """
        :type language: str
        :returns: list
        """
        import nltk
        from nltk.corpus import stopwords

        nltk.download("stopwords", quiet=True)
        if language in stopwords.fileids():
            logger.info("Loaded NLTK stopwords for %s", language)
            return stopwords.words(language)

        logger.warning("No NLTK stopwords for %s", language)
        return []

    @property
    def stopwords_list(self) -> List[str]:
        return self._stopwords_list



import re
from typing import Optional

from ftfy import TextFixerConfig, fix_text
from hyfi.composer.base import BaseConfig
from pydantic import BaseModel

from lexikanon import HyFI
from lexikanon.utils import (
    ACCENTS,
    APOSTROPHES,
    CONTROLS,
    DOUBLE_QUOTES,
    HYPHENS,
    LEFT_PARENTHESES,
    MINUSES,
    QUOTES,
    RIGHT_PARENTHESES,
    SINGLE_QUOTES,
    SLASHES,
    TILDES,
    emoticon_normalize,
    hanja_to_hangle,
)

logger = HyFI.getLogger(__name__)


class FyfyConfig(BaseModel):
    """
    unescape_html: true
    remove_terminal_escapes: true
    fix_encoding: true
    restore_byte_a0: true
    replace_lossy_sequences: true
    decode_inconsistent_utf8: true
    fix_c1_controls: true
    fix_latin_ligatures: true
    fix_character_width: true
    uncurl_quotes: true
    fix_line_breaks: true
    fix_surrogates: true
    remove_control_chars: true
    normalization: NFKC
    max_decode_length: 1000000
    """

    unescape_html: bool = True
    remove_terminal_escapes: bool = True
    fix_encoding: bool = True
    restore_byte_a0: bool = True
    replace_lossy_sequences: bool = True
    decode_inconsistent_utf8: bool = True
    fix_c1_controls: bool = True
    fix_latin_ligatures: bool = True
    fix_character_width: bool = True
    uncurl_quotes: bool = True
    fix_line_breaks: bool = True
    fix_surrogates: bool = True
    remove_control_chars: bool = True
    normalization: str = "NFKC"
    max_decode_length: int = 1000000


class SpacesConfig(BaseModel):
    """
    strip: true
    fix_whitespaces: true
    collapse_whitespaces: true
    replace_tabs: true
    num_spaces_for_tab: 4
    """

    strip: bool = True
    fix_whitespaces: bool = True
    collapse_whitespaces: bool = True
    replace_tabs: bool = True
    num_spaces_for_tab: int = 4


class SpecialCharactersConfig(BaseModel):
    """
    fix_hyphens: true
    fix_ellipsis: true
    fix_slashes: true
    fix_tildes: true
    fix_emoticons: false
    single_quotes_only: false
    regular_parentheses_only: false
    """

    fix_hyphens: bool = True
    fix_ellipsis: bool = True
    fix_slashes: bool = True
    fix_tildes: bool = True
    fix_emoticons: bool = False
    single_quotes_only: bool = False
    regular_parentheses_only: bool = False


class Normalizer(BaseConfig):
    """Main Normalizer class for generic text.
    Normalize unicode, hyphens, quotes, whitespace.
    By default, the normal form NFKC is used for unicode normalization.
    This applies a compatibility decomposition,
    under which equivalent characters are unified, followed by a canonical composition.
    See Python docs for information
    on normal forms: http://docs.python.org/2/library/unicodedata.html#unicodedata.normalize
    """

    ftfy: FyfyConfig = FyfyConfig()
    spaces: SpacesConfig = SpacesConfig()
    special_characters: SpecialCharactersConfig = SpecialCharactersConfig()
    hanja2hangle: bool = True
    num_repeats: int = 2

    _ftfy_cfg: Optional[TextFixerConfig] = None

    def __call__(self, text: Optional[str]) -> str:
        """Calling a normalizer instance like a function just calls the normalize method."""
        return self.normalize(text)

    def normalize(self, text: Optional[str]) -> str:
        """Run the Normalizer on a string.
        :param text: The string to normalize.
        """
        if text is None:
            return ""
        text = self._fix_text(text)

        # Normalize to canonical unicode (using NFKC by default)
        # if self.form is not None:
        #     text = unicodedata.normalize(self.form, text)

        if self.ftfy.remove_control_chars:
            text = self._remove_control_chars(text)

        # if self.fix_line_breaks:
        #     text = text.replace('\u2028', '\n').replace('\u2029', '\n').replace('\r\n', '\n').replace('\r', '\n')

        if self.special_characters.fix_hyphens:
            text = self._fix_hyphens(text)

        if self.ftfy.uncurl_quotes:
            text = self._uncurl_quotes(text)

        if self.special_characters.fix_ellipsis:
            text = self._fix_ellipsis(text)

        if self.special_characters.fix_slashes:
            text = self._fix_slashes(text)

        if self.special_characters.fix_tildes:
            text = self._fix_tildes(text)

        if self.spaces.replace_tabs:
            repacement_spaces = " " * self.spaces.num_spaces_for_tab
            text = self._replace_tabs(text, replacement_spaces=repacement_spaces)

        if self.spaces.fix_whitespaces:
            text = self._fix_whitespaces(text)

        if self.spaces.collapse_whitespaces:
            text = self._collapse_whitespaces(text)

        if self.spaces.strip and text is not None:
            text = text.strip()

        if self.special_characters.single_quotes_only:
            text = self._single_quotes_only(text)

        if self.special_characters.regular_parentheses_only:
            text = self._regular_parentheses_only(text)

        if self.hanja2hangle:
            text = self._hanja_to_hangle(text)

        if self.special_characters.fix_emoticons and text is not None:
            text = emoticon_normalize(text, num_repeats=self.num_repeats)

        return text or ""

    def _fix_text(self, text: Optional[str]) -> Optional[str]:
        if self._ftfy_cfg is None:
            self._ftfy_cfg = TextFixerConfig(**self.ftfy.model_dump())
        return None if text is None else fix_text(str(text), self._ftfy_cfg)

    def _remove_control_chars(self, text: Optional[str]) -> Optional[str]:
        """
        Strip out any control characters (they occasionally creep in somehow)
        """
        return self._replace_special_characters(text, CONTROLS, "")

    def _fix_hyphens(self, text: Optional[str]) -> Optional[str]:
        """
        Normalize all hyphens, minuses and dashes to ascii hyphen-minus and remove soft hyphen entirely
        """
        # TODO: Better normalization of em/en dashes to '--' if surrounded by spaces or start/end?
        if text is None:
            return None
        for hyphen in HYPHENS | MINUSES:
            text = text.replace(hyphen, "-")
        text = text.replace("\u00ad", "")
        return text

    def _uncurl_quotes(self, text: Optional[str]) -> Optional[str]:
        """
        Normalize all quotes and primes to ascii apostrophe and quotation mark
        """
        if text is None:
            return None
        for double_quote in DOUBLE_QUOTES:
            text = text.replace(double_quote, '"')  # \u0022
        for single_quote in SINGLE_QUOTES | APOSTROPHES | ACCENTS:
            text = text.replace(single_quote, "'")  # \u0027
        text = text.replace("′", "'")  # \u2032 prime
        text = text.replace("‵", "'")  # \u2035 reversed prime
        text = text.replace("″", "''")  # \u2033 double prime
        text = text.replace("‶", "''")  # \u2036 reversed double prime
        text = text.replace("‴", "'''")  # \u2034 triple prime
        text = text.replace("‷", "'''")  # \u2037 reversed triple prime
        text = text.replace("⁗", "''''")  # \u2057 quadruple prime
        return text

    def _fix_ellipsis(self, text: Optional[str]) -> Optional[str]:
        """
        Normalize ellipses to three full stops
        """
        if text is None:
            return None
        text = text.replace("…", "...").replace(" . . . ", " ... ")  # \u2026
        return text

    def _fix_slashes(self, text: Optional[str]) -> Optional[str]:
        """
        Normalize slash characters to ascii slash
        """
        return self._replace_special_characters(text, SLASHES, "/")

    def _fix_tildes(self, text: Optional[str]) -> Optional[str]:
        """
        Normalize tilde characters to ascii tilde
        """
        return self._replace_special_characters(text, TILDES, "~")

    def _replace_tabs(
        self, text: Optional[str], replacement_spaces=" " * 4
    ) -> Optional[str]:
        """
        Replace tabs with spaces
        """
        if text is None:
            return None
        text = text.replace("\t", replacement_spaces)
        return text

    def _fix_whitespaces(self, text: Optional[str]) -> Optional[str]:
        """
        Normalize unusual whitespace not caught by unicodedata
        """
        if text is None:
            return None
        text = text.replace("\u000b", " ").replace("\u000c", " ").replace("\u0085", " ")
        return text

    def _collapse_whitespaces(self, text: Optional[str]) -> Optional[str]:
        """
        Collapse all whitespace to a single space
        """
        if text is None:
            return None

        text = re.sub(r" +", " ", text)
        return text

    def _single_quotes_only(self, text: Optional[str]) -> Optional[str]:
        """
        Replace all double quotes with single quotes
        """
        return self._replace_special_characters(text, QUOTES, "'")

    def _replace_special_characters(
        self, text: Optional[str], special_chars, replacement
    ) -> Optional[str]:
        if text is None:
            return None
        for control in special_chars:
            text = text.replace(control, replacement)
        return text

    # Convert all brackets to regular parentheses
    def _regular_parentheses_only(self, text: Optional[str]) -> Optional[str]:
        """
        Replace all curly brackets with regular parentheses
        """
        if text is None:
            return None
        for ob in LEFT_PARENTHESES:
            text = text.replace(ob, "(")
        for cb in RIGHT_PARENTHESES:
            text = text.replace(cb, ")")
        return text

    def _hanja_to_hangle(self, text: Optional[str]) -> Optional[str]:
        """
        Convert all hanja to hangle
        """
        if text is None:
            return None
        text = hanja_to_hangle(text)
        return text

    def _fix_emoticons(self, text: Optional[str], num_repeats=2) -> Optional[str]:
        """
        Replace emoticons with their text equivalents
        """
        if text is None:
            return None
        text = emoticon_normalize(text, num_repeats=num_repeats)
        return text



from typing import List, Optional, Tuple

from pydantic import BaseModel

from lexikanon import HyFI
from lexikanon.tokenizers.base import Tokenizer

logger = HyFI.getLogger(__name__)


class MecabTagger(BaseModel):
    """
    userdic_path:
    backend: ekonlpy
    verbose: false
    """

    userdic_path: Optional[str] = None
    backend: str = "ekonlpy"
    verbose: bool = False

    def _parse(self, text: str) -> List[Tuple[str, str]]:
        from ekonlpy import Mecab

        mecab = Mecab()
        return mecab.pos(text)


class MecabTokenizer(Tokenizer):
    tagger: MecabTagger = MecabTagger()
    flatten: bool = True

    def parse(self, text: str) -> List[Tuple[str, str]]:
        return self.tagger._parse(text)



import itertools
import re
import unicodedata
from typing import Optional, Set

from lexikanon import HyFI

from .hangle import compose, decompose
from .hanja import translate as hanja2hangle

logger = HyFI.getLogger(__name__)


doublespace_pattern = re.compile(r"\s+")
repeatchars_pattern = re.compile(r"(\w)\\1{2,}")
number_pattern = re.compile(r"[0-9]")
punctuation_pattern = re.compile(r"[,\.\?\!]")
symbol_pattern = re.compile(r"[()\[\]\{\}`]")
hangle_pattern = re.compile(r"[ㄱ-ㅎㅏ-ㅣ가-힣]")
alphabet_pattern = re.compile(r"[a-zA-Z]")

hangle_filter = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣]")
hangle_number_filter = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣0-9]")
text_filter = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9,\.\?\!&·\"'\(\)\[\]\{\}+\-\\\/\*×%]")
# text_filter = re.compile(r'[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9,\.\?\!&·\"\'-()\[\]\{\}]')
text_filter_for_lrgraph = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9,\.\?\!&\"'-()\[\]\{\}]")


#: Control characters.
CONTROLS = {
    "\u0001",
    "\u0002",
    "\u0003",
    "\u0004",
    "\u0005",
    "\u0006",
    "\u0007",
    "\u0008",
    "\u000e",
    "\u000f",
    "\u0011",
    "\u0012",
    "\u0013",
    "\u0014",
    "\u0015",
    "\u0016",
    "\u0017",
    "\u0018",
    "\u0019",
    "\u001a",
    "\u001b",
}
# There are further control characters, but they are instead replaced with a space by unicode normalization
# '\u0009', '\u000a', '\u000b', '\u000c', '\u000d', '\u001c',  '\u001d', '\u001e', '\u001f'


#: Hyphen and dash characters.
HYPHENS = {
    "-",  # \u002d Hyphen-minus
    "‐",  # \u2010 Hyphen
    "‑",  # \u2011 Non-breaking hyphen
    "⁃",  # \u2043 Hyphen bullet
    "‒",  # \u2012 figure dash
    "–",  # \u2013 en dash
    "—",  # \u2014 em dash
    "―",  # \u2015 horizontal bar
}

#: Minus characters.
MINUSES = {
    "-",  # \u002d Hyphen-minus
    "−",  # \u2212 Minus
    "－",  # \uff0d Full-width Hyphen-minus
    "⁻",  # \u207b Superscript minus
}

#: Plus characters.
PLUSES = {
    "+",  # \u002b Plus
    "＋",  # \uff0b Full-width Plus
    "⁺",  # \u207a Superscript plus
}

#: Slash characters.
SLASHES = {
    "/",  # \u002f Solidus
    "⁄",  # \u2044 Fraction slash
    "∕",  # \u2215 Division slash
}

#: Tilde characters.
TILDES = {
    "~",  # \u007e Tilde
    "˜",  # \u02dc Small tilde
    "⁓",  # \u2053 Swung dash
    "∼",  # \u223c Tilde operator
    "∽",  # \u223d Reversed tilde
    "∿",  # \u223f Sine wave
    "〜",  # \u301c Wave dash
    "～",  # \uff5e Full-width tilde
}

#: Apostrophe characters.
APOSTROPHES = {
    "'",  # \u0027
    "’",  # \u2019
    "՚",  # \u055a
    "Ꞌ",  # \ua78b
    "ꞌ",  # \ua78c
    "＇",  # \uff07
}

#: Single quote characters.
SINGLE_QUOTES = {
    "'",  # \u0027
    "‘",  # \u2018
    "’",  # \u2019
    "‚",  # \u201a
    "‛",  # \u201b
}

#: Double quote characters.
DOUBLE_QUOTES = {
    '"',  # \u0022
    "“",  # \u201c
    "”",  # \u201d
    "„",  # \u201e
    "‟",  # \u201f
}

#: Accent characters.
ACCENTS = {
    "`",  # \u0060
    "´",  # \u00b4
}

#: Prime characters.
PRIMES = {
    "′",  # \u2032
    "″",  # \u2033
    "‴",  # \u2034
    "‵",  # \u2035
    "‶",  # \u2036
    "‷",  # \u2037
    "⁗",  # \u2057
}

#: Quote characters, including apostrophes, single quotes, double quotes, accents and primes.
QUOTES = APOSTROPHES | SINGLE_QUOTES | DOUBLE_QUOTES | ACCENTS | PRIMES

#: Uppercase and lowercase greek letters.
GREEK = {
    "Α",  # \u0391
    "Β",  # \u0392
    "Γ",  # \u0393
    "Δ",  # \u0394
    "Ε",  # \u0395
    "Ζ",  # \u0396
    "Η",  # \u0397
    "Θ",  # \u0398
    "Ι",  # \u0399
    "Κ",  # \u039a
    "Λ",  # \u039b
    "Μ",  # \u039c
    "Ν",  # \u039d
    "Ξ",  # \u039e
    "Ο",  # \u039f
    "Π",  # \u03a0
    "Ρ",  # \u03a1
    "Σ",  # \u03a3
    "Τ",  # \u03a4
    "Υ",  # \u03a5
    "Φ",  # \u03a6
    "Χ",  # \u03a7
    "Ψ",  # \u03a8
    "Ω",  # \u03a9
    "α",  # \u03b1
    "β",  # \u03b2
    "γ",  # \u03b3
    "δ",  # \u03b4
    "ε",  # \u03b5
    "ζ",  # \u03b6
    "η",  # \u03b7
    "θ",  # \u03b8
    "ι",  # \u03b9
    "κ",  # \u03ba
    "λ",  # \u03bb
    "μ",  # \u03bc
    "ν",  # \u03bd
    "ξ",  # \u03be
    "ο",  # \u03bf
    "π",  # \u03c0
    "ρ",  # \u03c1
    "σ",  # \u03c3
    "τ",  # \u03c4
    "υ",  # \u03c5
    "φ",  # \u03c6
    "χ",  # \u03c7
    "ψ",  # \u03c8
    "ω",  # \u03c9
}

#: Names of greek letters spelled out as words.
GREEK_WORDS = {
    "Alpha",
    "Beta",
    "Gamma",
    "Delta",
    "Epsilon",
    "Zeta",
    "Eta",
    "Theta",
    "Iota",
    "Kappa",
    "Lambda",
    "Mu",
    "Nu",
    "Xi",
    "Omicron",
    "Pi",
    "Rho",
    "Sigma",
    "Tau",
    "Upsilon",
    "Phi",
    "Chi",
    "Psi",
    "Omega",
    "alpha",
    "beta",
    "gamma",
    "delta",
    "epsilon",
    "zeta",
    "eta",
    "theta",
    "iota",
    "kappa",
    "lamda",
    "mu",
    "nu",
    "xi",
    "omicron",
    "pi",
    "rho",
    "sigma",
    "tau",
    "upsilon",
    "phi",
    "chi",
    "psi",
    "omega",
}

#: Words that should not be capitalized in titles.
SMALL = {
    "a",
    "an",
    "and",
    "as",
    "at",
    "but",
    "by",
    "en",
    "for",
    "if",
    "in",
    "of",
    "on",
    "or",
    "the",
    "to",
    "v",
    "via",
    "vs",
}

#: Words that should not be capitalized in names.
NAME_SMALL = {
    "abu",
    "bon",
    "bin",
    "da",
    "dal",
    "del",
    "der",
    "de",
    "di",
    "dí",
    "ibn",
    "la",
    "le",
    "san",
    "st",
    "ste",
    "van",
    "vel",
    "von",
    "y",
}

# This isn't every possible TLD, just the most common, to avoid false positives.
TLDS = {
    "aero",
    "asia",
    "biz",
    "cat",
    "com",
    "coop",
    "edu",
    "eu",
    "gov",
    "info",
    "int",
    "jobs",
    "mil",
    "mobi",
    "museum",
    "name",
    "net",
    "org",
    "pro",
    "tel",
    "travel",
    "xxx",
    "ad",
    "as",
    "ar",
    "au",
    "br",
    "bz",
    "ca",
    "cc",
    "cd",
    "co",
    "ch",
    "cn",
    "de",
    "dj",
    "es",
    "fr",
    "fm",
    "it",
    "io",
    "jp",
    "la",
    "ly",
    "me",
    "ms",
    "nl",
    "no",
    "nu",
    "ru",
    "sc",
    "se",
    "sr",
    "su",
    "tk",
    "tv",
    "uk",
    "us",
    "ws",
}

#: A variety of numbers, spelled out as words.
NUMBERS = {
    "zero",
    "one",
    "two",
    "three",
    "four",
    "five",
    "six",
    "seven",
    "eight",
    "nine",
    "ten",
    "eleven",
    "twelve",
    "thirteen",
    "fourteen",
    "fifteen",
    "sixteen",
    "seventeen",
    "eighteen",
    "nineteen",
    "twenty",
    "thirty",
    "forty",
    "fifty",
    "sixty",
    "seventy",
    "eighty",
    "ninety",
    "hundred",
    "thousand",
    "million",
    "billion",
    "trillion",
}
LEFT_PARENTHESES = {"(", "[", "{", "&lt;"}
RIGHT_PARENTHESES = {")", "]", "}", "&gt;"}
#: Regular expression that matches email addresses.
EMAIL_RE = re.compile(r"([\w\-\.\+%]+@(\w[\w\-]+\.)+[\w\-]+)", re.I | re.U)
#: Regular expression that matches DOIs.
DOI_RE = re.compile(r"^10\.\d{4,9}/[-\._;()/:A-Z0-9]+$", re.U)
#: Regular expression that matches ISSNs.
ISSN_RE = re.compile(r"^\d{4}-\d{3}[\dX]$", re.U)
#: Regular expression that matches control characters not allowed in XML.
CONTROL_RE = re.compile(
    "[^\u0020-\uD7FF\u0009\u000A\u000D\uE000-\uFFFD\u10000-\u10FFFF]+"
)


def levenshtein(s1: str, s2: str, allow_substring: bool = False) -> int:
    """Return the Levenshtein distance between two strings.
    The Levenshtein distance (a.k.a "edit difference") is the number of characters that need to be substituted,
    inserted or deleted to transform s1 into s2.
    Setting the `allow_substring` parameter to True allows s1 to be a
    substring of s2, so that, for example, "hello" and "hello there" would have a distance of zero.
    :param string s1: The first string
    :param string s2: The second string
    :param bool allow_substring: Whether to allow s1 to be a substring of s2
    :returns: Levenshtein distance.
    :rtype int
    """
    len1, len2 = len(s1), len(s2)
    lev = [[0] * (len2 + 1) for _ in range(len1 + 1)]
    for i in range(len1 + 1):
        lev[i][0] = i
    for j in range(len2 + 1):
        lev[0][j] = 0 if allow_substring else j
    for i, j in itertools.product(range(len1), range(len2)):
        lev[i + 1][j + 1] = min(
            lev[i][j + 1] + 1, lev[i + 1][j] + 1, lev[i][j] + (s1[i] != s2[j])
        )
    return min(lev[len1]) if allow_substring else lev[len1][len2]


def bracket_level(
    text: str, _open: Optional[Set[str]] = None, _close: Optional[Set[str]] = None
) -> int:
    """Return 0 if string contains balanced brackets or no brackets."""
    if _open is None:
        _open = {"(", "[", "{"}
    if _close is None:
        _close = {")", "]", "}"}
    level = 0
    for c in text:
        if c in _open:
            level += 1
        elif c in _close:
            level -= 1
    return level


def is_punct(text: str) -> bool:
    return all(unicodedata.category(char).startswith("P") for char in text)


def is_ascii(text: str) -> bool:
    return all(ord(char) < 128 for char in text)


def like_url(text: str) -> bool:
    if not text:
        return False
    if text.startswith("http://"):
        return True
    elif text.startswith("www.") and len(text) >= 5:
        return True
    if len(text) < 2 or text[0] == "." or text[-1] == "." or "." not in text:
        return False
    tld = text.rsplit(".", 1)[1].split(":", 1)[0]
    return True if tld.endswith("/") else bool(tld.isalpha() and tld in TLDS)


def like_number(text: str) -> bool:
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if like_number(num) and like_number(denom):
            return True
    return text in NUMBERS


def word_shape(text: str) -> str:
    prev_m = ""
    seq = 0
    shape = []
    for c in text:
        if c.isdigit():
            m = "d"  # Digits
        elif c in GREEK:
            m = "g"  # Greek letters
        elif c.isalpha():
            m = "X" if c.isupper() else "x"  # Uppercase or lowercase alphabetical
        elif c in QUOTES:
            m = "'"  # Quotes and apostrophes
        elif c in {":", ";"}:
            m = ":"  # Colons and semicolons
        elif c in {"!", "?", "."}:
            m = "."  # Sentence ends
        elif c in {"(", "[", "{", ")", "]", "}"}:
            m = "b"  # Brackets
        elif c in {"°", "%"}:
            m = "u"  # units
        elif c in {"■", "◼", "●", "▲", "○", "◆", "▼", "⧫", "△", "◇", "▽", "⬚", "□"}:
            m = "l"  # list markers
        elif c in {",", "$", "&", "-"}:
            m = c  # Stay the same
        else:
            m = "*"
            # Everything else, symbols etc:
            # {'=', '+', '*', '_', '|', '@', '×', '÷', '±', '<', '≤', '>', '≥', '≦', '≡', '≅', '≈', '≃', '≲',
            # '→', '←', '⇄', '≪', '≫', '↔', '≠', '∝', '∈', '⇌', '⇋', '⋯', '~', '·', '•', '√', '⊃', '∑', '∏',
            # '®', '∞', '∂', '∫', '∇', '∧', '⟨', '⟩'}
        if m == prev_m:
            seq += 1
        else:
            seq = 0
            prev_m = m
        if seq < 3:
            shape.append(m)
    return "".join(shape)


def remove_doublespace(sent: str) -> str:
    return doublespace_pattern.sub(" ", sent)


def repeat_normalize(sent: str, num_repeats: int = 2) -> str:
    if num_repeats > 0:
        sent = repeatchars_pattern.sub("\\1" * num_repeats, sent)
    sent = doublespace_pattern.sub(" ", sent)
    return sent.strip()


def emoticon_normalize(sent: str, num_repeats: int = 2) -> str:
    if not sent:
        return sent

    # Pattern matching ㅋ쿠ㅜ
    def pattern(idx):
        # Jaum: 0, Moum: 1, Complete: 2, else -1
        if 12593 <= idx <= 12622:
            return 0
        elif 12623 <= idx <= 12643:
            return 1
        elif 44032 <= idx <= 55203:
            return 2
        else:
            return -1

    idxs = [pattern(ord(c)) for c in sent]
    sent_ = []
    last_idx = len(idxs) - 1
    for i, (idx, c) in enumerate(zip(idxs, sent)):
        if (i > 0 and i < last_idx) and (
            idxs[i - 1] == 0 and idx == 2 and idxs[i + 1] == 1
        ):
            cho, jung, jong = decompose(c)
            if (cho == sent[i - 1]) and (jung == sent[i + 1]) and (jong == " "):
                sent_.append(cho)
                sent_.append(jung)
            else:
                sent_.append(c)
        elif (i < last_idx) and (idx == 2) and (idxs[i + 1] == 0):
            cho, jung, jong = decompose(c)
            if jong == sent[i + 1]:
                sent_.append(compose(cho, jung, " "))
                sent_.append(jong)
        elif (i > 0) and (idx == 2 and idxs[i - 1] == 0):
            cho, jung, jong = decompose(c)
            if cho == sent[i - 1]:
                sent_.append(cho)
                sent_.append(jung)
        else:
            sent_.append(c)
    return repeat_normalize("".join(sent_), num_repeats)


def hanja_to_hangle(sent: str) -> str:
    return hanja2hangle(sent, "substitution")


def only_hangle(sent: str) -> str:
    sent = hanja_to_hangle(sent)
    return doublespace_pattern.sub(" ", hangle_filter.sub(" ", sent)).strip()


def only_hangle_number(sent: str) -> str:
    sent = hanja_to_hangle(sent)
    return doublespace_pattern.sub(" ", hangle_number_filter.sub(" ", sent)).strip()


def only_text(sent: str) -> str:
    sent = hanja_to_hangle(sent)
    return doublespace_pattern.sub(" ", text_filter.sub(" ", sent)).strip()


def remain_hangle_on_last(eojeol: str) -> str:
    matchs = list(hangle_pattern.finditer(eojeol))
    if not matchs:
        return ""
    last_index = matchs[-1].span()[1]
    return eojeol[:last_index].strip()


def normalize_sent_for_lrgraph(sent: str) -> str:
    sent = hanja_to_hangle(sent)
    sent = text_filter_for_lrgraph.sub(" ", sent)
    sent = symbol_pattern.sub(" ", sent)
    sent_ = [remain_hangle_on_last(eojeol) for eojeol in sent.split()]
    sent_ = [eojeol for eojeol in sent_ if eojeol]
    return " ".join(sent_) if sent_ else ""
